﻿/*
#A part of NonVisual Desktop Access (NVDA)
#Copyright (C) 2010-2012 Masataka Shinke, Takuya Nishimoto (nvdajp)
#This file is covered by the GNU General Public License.
#See the file COPYING for more details.
#
# nvdajpimeRPC/tsf.cpp
# 
*/

#include "nvdajpime.h"
#include "tls.h"
#include "tsf.h"
#include "debugout.h"

CnvdajpimeTSF::CnvdajpimeTSF()
{
    DebugOut(2, L"RPC: CnvdajpimeTSF constructed.");
    _dwThreadMgrEventSinkCookie = TF_INVALID_COOKIE;
    _dwTextEditSinkCookie = TF_INVALID_COOKIE;
    _pTextEditSinkContext = NULL;
    _pThreadMgr = NULL;
    _pszCompositionText = NULL;
    _uCompositionText = 0;

    _cRef = 1;
}

CnvdajpimeTSF::~CnvdajpimeTSF()
{
    DebugOut(2, L"RPC: CnvdajpimeTSF destructed.");
}

// IUnknown
STDAPI CnvdajpimeTSF::QueryInterface(REFIID riid, void **ppvObj)
{
    if (ppvObj == NULL)
        return E_INVALIDARG;

    *ppvObj = NULL;

    if (IsEqualIID(riid, IID_IUnknown) ||
        IsEqualIID(riid, IID_ITfTextEditSink))
    {
        *ppvObj = (ITfTextEditSink *)this;
    }
    else if (IsEqualIID(riid, IID_ITfThreadMgrEventSink))
    {
        *ppvObj = (ITfThreadMgrEventSink *)this;
    }

    if (*ppvObj)
    {
        AddRef();
        return S_OK;
    }

    return E_NOINTERFACE;
}

// IUnknown
STDAPI_(ULONG) CnvdajpimeTSF::AddRef()
{
    DebugOut(2, L"RPC: CnvdajpimeTSF::AddRef() called.");
    return ++_cRef;
}

// IUnknown
STDAPI_(ULONG) CnvdajpimeTSF::Release()
{
    LONG cr = --_cRef;
    assert(_cRef >= 0);
    if (_cRef == 0)
    {
        delete this;
        DebugOut(2, L"RPC: CnvdajpimeTSF::Release() deleted.");
    }
    return cr;
}

// static method
BOOL CnvdajpimeTSF::_Init()
{
    // DebugOut(2, L"RPC: CnvdajpimeTSF::_Init() called.");
    TLS *ptls = TLS::GetTLS();

    if (!ptls)
        return FALSE;

    if (ptls->_pReadCompTSF)
        return TRUE;

	ptls->_pReadCompTSF = new CnvdajpimeTSF;
    if (!ptls->_pReadCompTSF)
        return FALSE;
    DebugOut(2, L"RPC: CnvdajpimeTSF::_Init() _pReadCompTSF created.");

    ITfThreadMgr *ptim;
    HRESULT hr;

    hr = CoCreateInstance(CLSID_TF_ThreadMgr, 
                          NULL, 
                          CLSCTX_INPROC_SERVER, 
                          IID_ITfThreadMgr, 
                          (void **)&ptim);

    if (hr != S_OK)
    {
        ptls->_pReadCompTSF->Release();
        ptls->_pReadCompTSF = NULL;
        return FALSE;
    }

    ptls->_pReadCompTSF->_pThreadMgr = ptim;

    ptls->_pReadCompTSF->_InitThreadMgrSink();

    ITfDocumentMgr *pDocMgr;
    if (SUCCEEDED(ptim->GetFocus(&pDocMgr)))
    {
        if (pDocMgr)
        {
           ptls->_pReadCompTSF->_InitTextEditSink(pDocMgr);
           pDocMgr->Release();
        }
    }
    return TRUE;
}

// ITfThreadMgrEventSink
STDAPI CnvdajpimeTSF::OnInitDocumentMgr(ITfDocumentMgr *pDocMgr)
{
    DebugOut(2, L"RPC: CnvdajpimeTSF::OnInitDocumentMgr() called.");
    return S_OK;
}

// ITfThreadMgrEventSink
STDAPI CnvdajpimeTSF::OnUninitDocumentMgr(ITfDocumentMgr *pDocMgr)
{
    DebugOut(2, L"RPC: CnvdajpimeTSF::OnUninitDocumentMgr() called.");
    return S_OK;
}

// ITfThreadMgrEventSink
STDAPI CnvdajpimeTSF::OnSetFocus(ITfDocumentMgr *pDocMgrFocus, ITfDocumentMgr *pDocMgrPrevFocus)
{
    if (!pDocMgrFocus)
    {
        DebugOut(2, L"RPC: CnvdajpimeTSF::OnSetFocus(NULL) called.");
    }
    return S_OK;
}

// ITfThreadMgrEventSink
STDAPI CnvdajpimeTSF::OnPushContext(ITfContext *pContext)
{
    //DebugOut(2, L"RPC: CnvdajpimeTSF::OnPushContext() called.");
    return S_OK;
}

// ITfThreadMgrEventSink
STDAPI CnvdajpimeTSF::OnPopContext(ITfContext *pContext)
{
    //DebugOut(2, L"RPC: CnvdajpimeTSF::OnPopContext() called.");
    return S_OK;
}

BOOL CnvdajpimeTSF::_InitThreadMgrSink()
{
    DebugOut(2, L"RPC: CnvdajpimeTSF::_InitThreadMgrSink() called.");
    ITfSource *pSource = NULL;
    BOOL fRet = FALSE;

    if (_pThreadMgr->QueryInterface(IID_ITfSource, (void **)&pSource) != S_OK)
        return FALSE;

    if (pSource->AdviseSink(IID_ITfThreadMgrEventSink, (ITfThreadMgrEventSink *)this, &_dwThreadMgrEventSinkCookie) != S_OK)
    {
        _dwThreadMgrEventSinkCookie = TF_INVALID_COOKIE;
        goto Exit;
    }
    fRet = TRUE;

Exit:
    pSource->Release();
    return fRet;
}

void CnvdajpimeTSF::_UninitThreadMgrSink()
{
    DebugOut(2, L"RPC: CnvdajpimeTSF::_UninitThreadMgrSink() called.");
    ITfSource *pSource = NULL;

    if (_dwThreadMgrEventSinkCookie == TF_INVALID_COOKIE)
        return;

    if (_pThreadMgr->QueryInterface(IID_ITfSource, (void **)&pSource) == S_OK)
    {
        pSource->UnadviseSink(_dwThreadMgrEventSinkCookie);
        pSource->Release();
    }

    _dwThreadMgrEventSinkCookie = TF_INVALID_COOKIE;
}

// ITfTextEditSink
STDAPI CnvdajpimeTSF::OnEndEdit(ITfContext *pContext, TfEditCookie ecReadOnly, ITfEditRecord *pEditRecord)
{
	DebugOut(2, L"RPC: CnvdajpimeTSF::OnEndEdit() Cookie:", ecReadOnly);
	BOOL fChanged = FALSE;
	if (pEditRecord->GetSelectionStatus(&fChanged) == S_OK && fChanged) {
		DebugOut(2, L"RPC: CnvdajpimeTSF::OnEndEdit() SelectionStatus:Changed");
		FlushSharedMemory();
		if (GetSharedMemory()->fReadCompRunning) {
			_CheckComposition(pContext, ecReadOnly);
			if (nvdajpimeRPC_testIfRunning()) {
				RPC_Bind();
				UINT cursorPos = 0;
				UINT deltaStart = 0;
				const wchar_t* composition = L"";
				const wchar_t* compAttr = L"";
				const wchar_t* result = L"";
				const wchar_t* textService = L"";
				if (_pszCompositionText != NULL) {
					textService = _pszCompositionText;
				}
				nvdajpimeRPC_inputMethodCallback(cursorPos, deltaStart, composition, compAttr, result, textService);
				RPC_Free();
			}
		}
	}
    return S_OK;
}

// if pDocMgr==NULL, just clear the previous sink
BOOL CnvdajpimeTSF::_InitTextEditSink(ITfDocumentMgr *pDocMgr)
{
    DebugOut(2, L"RPC: CnvdajpimeTSF::_InitTextEditSink() called.");
    ITfSource *pSource = NULL;

    if (_dwTextEditSinkCookie != TF_INVALID_COOKIE)
    {
        if (_pTextEditSinkContext->QueryInterface(IID_ITfSource, (void **)&pSource) == S_OK)
        {
            pSource->UnadviseSink(_dwTextEditSinkCookie);
            pSource->Release();
            DebugOut(2, L"RPC: CnvdajpimeTSF::_InitTextEditSink() ITfSource::UnadviseSink() done.");
        }

        _pTextEditSinkContext->Release();
        _pTextEditSinkContext = NULL;
        _dwTextEditSinkCookie = TF_INVALID_COOKIE;
    }

    if (pDocMgr == NULL)
        return TRUE;

    if (pDocMgr->GetTop(&_pTextEditSinkContext) != S_OK)
        return FALSE;

    if (_pTextEditSinkContext == NULL)
        return TRUE; 

    BOOL fRet = FALSE;
    if (_pTextEditSinkContext->QueryInterface(IID_ITfSource, (void **)&pSource) == S_OK)
    {
        if (pSource->AdviseSink(IID_ITfTextEditSink, (ITfTextEditSink *)this, &_dwTextEditSinkCookie) == S_OK)
        {
            DebugOut(2, L"RPC: CnvdajpimeTSF::_InitTextEditSink() ITfSource::AdviseSink() done.");
            fRet = TRUE;
        }
        else
        {
            _dwTextEditSinkCookie = TF_INVALID_COOKIE;
        }
        pSource->Release();
    }

    if (fRet == FALSE)
    {
        _pTextEditSinkContext->Release();
        _pTextEditSinkContext = NULL;
    }

    return fRet;
}

void CnvdajpimeTSF::_AppendCompositionText(ITfRange *pRange, TfEditCookie ecReadOnly)
{
    BOOL fEmpty = FALSE;
    while (pRange->IsEmpty(ecReadOnly, &fEmpty) == S_OK && !fEmpty)
    {
        WCHAR *pwstr;
        WCHAR wstr[256 + 1];
        ULONG ulcch = ARRAYSIZE(wstr) - 1;
        ULONG ulcchTotal;
        pRange->GetText(ecReadOnly, TF_TF_MOVESTART | TF_TF_IGNOREEND, wstr, ulcch, &ulcch);

        ulcchTotal = _uCompositionText + ulcch;
        if (_pszCompositionText)
            pwstr = (WCHAR *)LocalReAlloc(_pszCompositionText, LPTR, (ulcchTotal + 1) * sizeof(WCHAR));
        else
            pwstr = (WCHAR *)LocalAlloc(LPTR, (ulcchTotal + 1) * sizeof(WCHAR));

        if (pwstr)
        {
            memcpy(pwstr+_uCompositionText, wstr,  ulcch * sizeof(WCHAR));
            _pszCompositionText = pwstr;
            _uCompositionText = ulcchTotal;
        }
    }
	if (nvdajpimeRPC_testIfRunning()) {
		RPC_Bind();
		UINT cursorPos = 0;
		UINT deltaStart = 0;
		const wchar_t* composition = L"";
		const wchar_t* compAttr = L"";
		const wchar_t* result = L"";
		const wchar_t* textService = L"";
		if (_pszCompositionText != NULL) {
			textService = _pszCompositionText;
		}
		nvdajpimeRPC_inputMethodCallback(cursorPos, deltaStart, composition, compAttr, result, textService);
		RPC_Free();
	}
	DebugOut(2, L"RPC: CnvdajpimeTSF::_Append() ", _pszCompositionText);
}

void CnvdajpimeTSF::_ClearCompositionText()
{
    DebugOut(2, L"RPC: CnvdajpimeTSF::_ClearCompositionText() called.");
    if (_pszCompositionText)
    {
        LocalFree(_pszCompositionText);
        _pszCompositionText = NULL;
    }
    _uCompositionText = 0;
}

BOOL CnvdajpimeTSF::_CheckComposition(ITfContext *pContext, TfEditCookie ecReadOnly)
{
    DebugOut(2, L"RPC: CnvdajpimeTSF::_CheckComposition() called.");
    HRESULT hr;
    ITfContextComposition *pContextComposition;
    WCHAR *pszSavedCompositionText = NULL;
    UINT  uSavedCompositionText = 0;

    if (_pszCompositionText)
    {
        pszSavedCompositionText = (WCHAR *)LocalAlloc(LPTR, (_uCompositionText + 1) * sizeof(WCHAR));
        if (!pszSavedCompositionText)
            return FALSE;

        memcpy(pszSavedCompositionText, 
               _pszCompositionText,
               _uCompositionText);

       uSavedCompositionText = _uCompositionText;
    }

    _ClearCompositionText();

    hr = pContext->QueryInterface(IID_ITfContextComposition,
                                  (void **)&pContextComposition);
    if (hr == S_OK)
    {
		IEnumITfCompositionView *pEnumCompositionView;

        hr = pContextComposition->EnumCompositions(&pEnumCompositionView);
        if (hr == S_OK)
        {
            ITfCompositionView *pCompositionView;
			
            while (pEnumCompositionView->Next(1, &pCompositionView, NULL) == S_OK)
            {
                ITfRange *pRange;
                hr = pCompositionView->GetRange(&pRange);
                if (hr == S_OK)
                {
                    LONG xx = _GetLength(pRange, ecReadOnly);
					_AppendCompositionText(pRange, ecReadOnly);
                    pRange->Release();
                }
                pCompositionView->Release();
            }
            pEnumCompositionView->Release();
        }
        pContextComposition->Release();
    }

    if (!_uCompositionText && _pszCompositionText)
    {
        _pszCompositionText[0] = L'\0';
    }

    if ((uSavedCompositionText != _uCompositionText) ||
        memcmp(pszSavedCompositionText, _pszCompositionText, _uCompositionText * sizeof(WCHAR)))
    {
		if (nvdajpimeRPC_testIfRunning()) {
			RPC_Bind();
			UINT cursorPos = 0;
			UINT deltaStart = 0;
			const wchar_t* composition = L"";
			const wchar_t* compAttr = L"";
			const wchar_t* result = L"";
			const wchar_t* textService = L"";
			if (_pszCompositionText != NULL) {
				textService = _pszCompositionText;
			}
			nvdajpimeRPC_inputMethodCallback(cursorPos, deltaStart, composition, compAttr, result, textService);
			RPC_Free();
		}
	}

    if (pszSavedCompositionText)
        LocalFree(pszSavedCompositionText);
	
	return TRUE;
}

LONG CnvdajpimeTSF::_GetLength(ITfRange *pRange, TfEditCookie ecReadOnly)
{
   LONG cch = 0;

   ITfRange *prangeT;
   TF_HALTCOND hc;

   if (pRange->Clone(&prangeT) == S_OK)
   {
       hc.pHaltRange = pRange;
       hc.aHaltPos = TF_ANCHOR_END;
       //hc.aHaltPos = TF_ANCHOR_START;
       hc.dwFlags = 0;

       prangeT->ShiftStart(ecReadOnly, LONG_MAX, &cch, &hc);
	   //prangeT->ShiftEnd(ecReadOnly, LONG_MAX, &cch, &hc);
       prangeT->Release();
   }
   
   DebugOut(2, L"RPC: CnvdajpimeTSF::_GetLength() returns ", cch);
   return cch;
}
