# coding: UTF-8
#brailleDisplayDrivers/kgs3.py
#A part of NonVisual Desktop Access (NVDA)
#This file is covered by the GNU General Public License.
#See the file COPYING for more details.
#Copyright (C) 2011-2012 Takuya Nishimoto, Masataka Shinke

import braille
import ui
import queueHandler
import inputCore
from logHandler import log
import config
from collections import OrderedDict
import hwPortUtils
import _kgs

def _exec(names, routingIndex):
	try:
		inputCore.manager.executeGesture(InputGesture(names, routingIndex))
	except:
		pass

loadCount = 0

class BrailleDisplayDriver(braille.BrailleDisplayDriver):
	name = "kgs3"
	description = _(u"KGS Driver 3")

	@classmethod
	def check(cls):
		return True		

	@classmethod
	def getPossiblePorts(cls):
		ar = [cls.AUTOMATIC_PORT]
		ar.extend([ (p["port"], p["friendlyName"]) for p in hwPortUtils.listComPorts() ])
		return OrderedDict(ar)

	def __init__(self, port="auto"):
		global loadCount
		log.info("connecting %s" % port)
		if loadCount:
			loadCount += 1
			_kgs.disconnect()
		else:
			if not _kgs.load():
				log.info("failed %s" % port)
				return
			loadCount = 1
			
		# comPort: auto = None, COM1 = 0, COM2 = 1, ...
		comPort = None
		if port[:3] == "COM":
			comPort = int(port[3:]) - 1
		if _kgs.connect(comPort):
			_kgs.setGestureHandler(lambda names, routingIndex: _exec(names, routingIndex))
			log.info("connected %s" % port)
			return

	def terminate(self):
		global loadCount
		super(BrailleDisplayDriver, self).terminate()
		loadCount -= 1
		if not loadCount:
			log.info("unloading")
			_kgs.unload()
		log.info("terminated")

	def _get_numCells(self):
		return _kgs.numCells
	
	def display(self, cells):
		log.debug(" ".join(("%02x" % c) for c in cells))
		_kgs.display(cells)

	gestureMap = inputCore.GlobalGestureMap({
		"globalCommands.GlobalCommands": {
			"kb:alt": ("br(kgs):alt",),
			"kb:escape": ("br(kgs):esc",),
			"kb:delete": ("br(kgs):del",),
			"kb:backspace": ("br(kgs):bs",),
			"kb:space": ("br(kgs):space",),
			"kb:enter": ("br(kgs):enter","br(kgs):ok","br(kgs):set","br(kgs):select",),
			"kb:control": ("br(kgs):ctrl",),
			"kb:upArrow": ("br(kgs):upArrow",),
			"kb:downArrow": ("br(kgs):downArrow",),
			"kb:leftArrow": ("br(kgs):leftArrow",),
			"kb:rightArrow": ("br(kgs):rightArrow",),
			"braille_routeTo": ("br(kgs):route",),
			"braille_scrollBack": ("br(kgs):func1","br(kgs):func3+leftArrow",),
			"braille_scrollForward": ("br(kgs):func4","br(kgs):func3+rightArrow",),
			"braille_previousLine": ("br(kgs):func3+upArrow",),
			"braille_nextLine": ("br(kgs):func3+downArrow",),
			"showGui": ("br(kgs):ins",),
			"sayAll": ("br(kgs):read",),
			"kb:a": ("br(kgs):dot1",),
			"kb:b": ("br(kgs):dot1+dot2",),
			"kb:c": ("br(kgs):dot1+dot4",),
			"kb:d": ("br(kgs):dot1+dot4+dot5",),
			"kb:e": ("br(kgs):dot1+dot5",),
			"kb:f": ("br(kgs):dot1+dot2+dot4",),
			"kb:g": ("br(kgs):dot1+dot2+dot4+dot5",),
			"kb:h": ("br(kgs):dot1+dot2+dot5",),
			"kb:i": ("br(kgs):dot2+dot4",),
			"kb:j": ("br(kgs):dot2+dot4+dot5",),
			"kb:k": ("br(kgs):dot1+dot3",),
			"kb:l": ("br(kgs):dot1+dot2+dot3",),
			"kb:m": ("br(kgs):dot1+dot3+dot4",),
			"kb:n": ("br(kgs):dot1+dot3+dot4+dot5",),
			"kb:o": ("br(kgs):dot1+dot3+dot5",),
			"kb:p": ("br(kgs):dot1+dot2+dot3+dot4",),
			"kb:q": ("br(kgs):dot1+dot2+dot3+dot4+dot5",),
			"kb:r": ("br(kgs):dot1+dot2+dot3+dot5",),
			"kb:s": ("br(kgs):dot2+dot3+dot4",),
			"kb:t": ("br(kgs):dot2+dot3+dot4+dot5",),
			"kb:u": ("br(kgs):dot1+dot3+dot6",),
			"kb:v": ("br(kgs):dot1+dot2+dot3+dot6",),
			"kb:w": ("br(kgs):dot2+dot4+dot5+dot6",),
			"kb:x": ("br(kgs):dot1+dot3+dot4+dot6",),
			"kb:y": ("br(kgs):dot1+dot3+dot4+dot5+dot6",),
			"kb:z": ("br(kgs):dot1+dot3+dot5+dot6",),
			"kb:.": ("br(kgs):dot2+dot5+dot6",),
			"kb::": ("br(kgs):dot2+dot5",),
			"kb:;": ("br(kgs):dot2+dot3",),
			"kb:,": ("br(kgs):dot2",),
			"kb:-": ("br(kgs):dot3+dot6",),
			"kb:?": ("br(kgs):dot2+dot3+dot6",),
			"kb:!": ("br(kgs):dot2+dot3+dot5",),
			"kb:'": ("br(kgs):dot3",),
		}
	})

class InputGesture(braille.BrailleDisplayGesture):

	source = BrailleDisplayDriver.name

	def __init__(self, names, routingIndex):
		super(InputGesture, self).__init__()
		self.id = "+".join(names)
		self.routingIndex = routingIndex
