#nvdajp_keyEvents.py
# coding: UTF-8
#A part of NonVisual Desktop Access (NVDA)
# Masamitsu Misono, Masataka.Shinke, Takuya Nishimoto
#
# input method support depends on
# nvdaHelper/{nvdajpime, nvdajpimeLoader, nvdajpimeRPC}
# and source/NVDAObjects/IAccessible/nvdajpime.py.
# debug messages within C++ code is available if non-zero numbers are given:
# \HKEY_CURRENT_USER\Software\nvdajp -> DebugIME64 (REG_DWORD) 
# \HKEY_CURRENT_USER\Software\nvdajp -> DebugIME32 (REG_DWORD) 
# DebugView or TraceSpy can be used. http://tracespy.codeplex.com/
#
import time
import languageHandler
import characterProcessing
import nvdajp_dic
import config
import ui
import queueHandler
from logHandler import log
import threading
import tones
import speech
import braille
import api
import win32con
import winUser
from ctypes import *
import win32gui
import win32process

lib = None
imc = None
WM_IME_COMPOSITION = 0x010F
WM_IME_CONTROL = 0x0283
IME_ON = 1
IME_OFF = 0
IMC_GETOPENSTATUS = 0x0005
IMC_SETOPENSTATUS = 0x0006
IMC_GETCONVERSIONMODE = 0x0001
IMC_SETCONVERSIONMODE = 0x0002

_prevCompositionContent = None
_lastKeyCode = None
_prevReportCandidate = None

def isSleepMode():
	app = api.getFocusObject().appModule
	if app and app.sleepMode: return True
	return False

_lastName = None
_lastIsMSAA = None
_lastAccID = None

# from InputMethods 5236 source/NVDAObjects/IAccessible/mscandui.py
def _get_description(symbols):
	descriptions=[]
	for symbol in symbols:
		try:
			symbolDescriptions=characterProcessing.getCharacterDescription(speech.getCurrentLanguage(),symbol)
		except TypeError:
			symbolDescriptions=None
		if symbolDescriptions:
			if len(symbols)>1:
				descriptions.append(_("{symbol} as in {description}").format(symbol=symbol,description=symbolDescriptions[0]))
			else:
				descriptions.extend(symbolDescriptions)
	if descriptions:
		return ", ".join(descriptions)
	return ""
	
def getDiscriminantReading(name):
	if nvdajp_dic.isJapaneseLocale(languageHandler.getLanguage()):
		return nvdajp_dic.getJapaneseDiscriminantReading(name)
	return _get_description(name)

def _speakText(s):
	speech.speakText(s, symbolLevel=characterProcessing.SYMLVL_ALL)

def _brailleMessage(s):
	try:
		braille.handler.message(s)
	except:
		pass

def _uiMessage(s):
	_speakText(s)
	_brailleMessage(s)

def reportCandidate(name, accid = None, isMSAA = False):
	"""NVDAObjects/IAccessible/nvdajpime also calls this method. """
	global _lastName, _lastIsMSAA, _lastAccID
	global _prevReportCandidate
	if _lastName == name:
		if (not _lastIsMSAA) and isMSAA:
			if accid:
				# already reported without accid, so add the accid.
				_speakText(_("candidate %d") % accid)
			_lastIsMSAA = isMSAA
			_lastAccID = accid
			return
		if _lastIsMSAA and (not isMSAA):
			_lastIsMSAA = isMSAA
			_lastAccID = accid
			return
		if (not _lastIsMSAA) and (not isMSAA) and (_lastAccID == accid):
			return
	cand = getDiscriminantReading(name)
	if accid:
		mes = cand + ' ' + _("candidate %d") % accid
	else:
		mes = cand
	if mes and (cand != _prevReportCandidate):
		speech.cancelSpeech()
		_uiMessage(mes)
	_prevReportCandidate = cand
	_lastName = name
	_lastIsMSAA = isMSAA
	_lastAccID = accid

def getDiff(prev, curr):
	if curr is None: return None
	if prev is None: return curr
	if prev == curr: return None
	m = min(len(prev), len(curr))
	for pos in xrange(m):
		if prev[pos] != curr[pos]:
			d = curr[pos:]
			return d
	d = curr[m:]
	return d

def candidateUpdated(curr):
	if not curr or isSleepMode(): return
	log.io('candidate:%s' % curr)
	queueHandler.queueFunction(queueHandler.eventQueue,reportCandidate, curr)

def compositionCleared(s = None):
	if isSleepMode(): return
	global _prevCompositionContent
	# Reading at IME session end 
	if config.conf["keyboard"]["nvdajpEnableKeyEvents"]:
		if s:
			queueHandler.queueFunction(queueHandler.eventQueue,_uiMessage, s)
		elif _prevCompositionContent:
			queueHandler.queueFunction(queueHandler.eventQueue,_uiMessage, _prevCompositionContent)
		_prevCompositionContent = None
	global _lastName, _lastIsMSAA, _lastAccID
	_lastName = None
	_lastIsMSAA = None
	_lastAccID = None
	global _prevReportCandidate
	_prevReportCandidate = None

def _getLastConversionStatus():		
	"""Input method enabled: 1. If disabled: 0. If unknown: None."""
	global imc
	ret = None
	hwnd = winUser.getForegroundWindow()
	if hwnd == 0: return None
	AttachThreadInput = windll.user32.AttachThreadInput
	if AttachThreadInput( 
		win32process.GetWindowThreadProcessId(hwnd)[0], 
		windll.kernel32.GetCurrentThreadId(), 
		True 
	) :
		hwnd = win32gui.GetFocus()
		if hwnd == 0: return None
		imc = windll.imm32.ImmGetDefaultIMEWnd(hwnd)
		if imc == 0: return None
		ret = winUser.sendMessage(imc, WM_IME_CONTROL, IMC_GETOPENSTATUS, 0)
		AttachThreadInput( 
			win32process.GetWindowThreadProcessId(hwnd)[0], 
			windll.kernel32.GetCurrentThreadId(), 
			False 
		)
	return ret

def _reportStopConversion():
	if isSleepMode(): return
	# Beep if IME mode is changed 
	if config.conf["keyboard"]["nvdajpImeBeep"]:
		tones.beep(440,20) 
	else:
		queueHandler.queueFunction(queueHandler.eventQueue,_uiMessage,_("stop input method"))

def _reportStartConversion():
	if isSleepMode(): return
	# Beep if IME mode is changed 
	if config.conf["keyboard"]["nvdajpImeBeep"]:
		tones.beep(880,20) 
	else:
		queueHandler.queueFunction(queueHandler.eventQueue,_uiMessage,_("start input method"))

def _reportToggleConversion():
	if isSleepMode(): return
	# Beep if IME mode is changed 
	if config.conf["keyboard"]["nvdajpImeBeep"]:
		tones.beep(784,20) 
	else:
		if not config.conf["keyboard"]["speakCommandKeys"]:
			queueHandler.queueFunction(queueHandler.eventQueue,_uiMessage,_("toggle input method"))

CONVERSIONMODE_DIC = {
	 0: _("half alphanumeric"),
	 3: _("half katakana"),
	 8: _("alphanumeric"),
	 9: _("hiragana"),
	11: _("katakana"),
	16: _("half alphanumeric"),
	19: _("half katakana roman"),
	24: _("alphanumeric"),
	25: _("hiragana roman"),
	27: _("katakana roman"),
} 

def _reportConversionMode():
	global imc
	if isSleepMode(): return
	if not imc: return
	m = winUser.sendMessage(imc, WM_IME_CONTROL, IMC_GETCONVERSIONMODE, 0)
	if m in CONVERSIONMODE_DIC:
		queueHandler.queueFunction(queueHandler.eventQueue,_uiMessage,CONVERSIONMODE_DIC[m])

# from keyboardHandler.internal_keyDownEvent
def reportKeyDownEvent(gesture):
	global _lastKeyCode
	global prevComposition, prevTextService
	_lastKeyCode = gesture.vkCode
	# TSF application do not notify input session end,
	# so we raise 'result string' event here.
	if _lastKeyCode == win32con.VK_RETURN and prevTextService:
		log.io("KD C:%s" % prevTextService)
		filterInputEvent('', '', '', '', prevTextService, 0) # result event
		prevTextService = None
		prevComposition = None
		return True
	# report when input method is enabled or disabled.
	key = u"+".join(gesture._keyNamesInDisplayOrder)
	if key in (
		"alt+IMEChangeStatus1",
		"IMEChangeStatus2",
		"IMEChangeStatus3",
		"alt+IMEChangeStatus2",
		"alt+IMEChangeStatus3"):
		log.io(key)
		ime = _getLastConversionStatus()
		if ime is None:
 			_reportToggleConversion()
		elif ime == 0:
			_reportStartConversion()
			_reportConversionMode()
		elif ime == 1:
			_reportStopConversion()
	return True

def deltaUpdated(s):
	if not s or isSleepMode(): return
	s = ' '.join([nvdajp_dic.get_short_desc(c) for c in s])
	queueHandler.queueFunction(queueHandler.eventQueue,_uiMessage, s)

def selectionUpdated(s):
	if not s or isSleepMode(): return
	s = getDiscriminantReading(s)
	queueHandler.queueFunction(queueHandler.eventQueue,_uiMessage, s)

#def caretCharUpdated(c):
#	if not c or isSleepMode(): return
#	queueHandler.queueFunction(queueHandler.eventQueue,_uiMessage, c)

prevInputTime = None
prevComposition = None
prevDelta = None
prevSelectedStr = None
prevCursorStr = None
prevResult = None
prevCursorPos = -1

def filterInputEvent(composition, delta, selectedStr, cursorStr, result, cursorPos):
	global prevInputTime
	global prevComposition, prevDelta, prevSelectedStr
	global prevCursorStr, prevResult, prevCursorPos

	sameTimeEvent = False
	if prevInputTime and time.time() - prevInputTime < 0.1:
		sameTimeEvent = True

	if result and not (result == prevResult and sameTimeEvent):
		compositionCleared(result)
	#elif cursorStr and not (cursorStr == prevCursorStr and sameTimeEvent):
	#	caretCharUpdated(cursorStr)
	elif selectedStr and not (selectedStr == prevSelectedStr and sameTimeEvent):
		selectionUpdated(selectedStr)
	elif delta and not (delta == prevDelta and sameTimeEvent):
		deltaUpdated(delta)

	prevInputTime = time.time()
	prevComposition = composition
	prevDelta = delta
	prevSelectedStr = selectedStr
	prevCursorStr = cursorStr
	prevResult = result
	prevCursorPos = cursorPos

prevTextServiceTime = None
prevTextService = None

# callback
# @param cursorPos position of caret. 0 is the left boundary of pre-edit or composition.
# @param deltaStart position of starting 'delta' within pre-edit or composition.
# @param composition both pre-edit and composition string can be obtained.
# @param compAttr attribute values for each characters, such as L"222221111000". 0: not converted, 1: selected, 2: not selected.
# @param result available if an IME session is just finished.
# @param textService composition text, available at ITfTextEditSink::OnEndEdit event. 
def inputMethodCallback(cursorPos, deltaStart, composition, compAttr, result, textService):
	global prevTextService, prevTextServiceTime
	if not composition and not result and not textService: return
	delta = ''
	cursorStr = ''
	selectedStr = ''
	if textService:
		# ignore if TSF update event is duplicated.
		if textService == prevTextService and prevTextServiceTime and time.time() - prevTextServiceTime < 0.1:
			return
		sameTimeEvent = True
		composition = textService
		# if reading string and first candidate is same, announce composition itself.
		if _lastKeyCode == win32con.VK_SPACE:
			delta = composition
		else:
			delta = getDiff(prevTextService, textService)
		log.io("TS C:%s D:%s S:%s M:%s R:%s P:%d" % (composition, delta, selectedStr, cursorStr, result, cursorPos))
		prevTextService = textService
		prevTextServiceTime = time.time()
	else:
		selectedStr = ''
		for pos in xrange(len(compAttr)):
			if compAttr[pos] == '1':
				selectedStr += composition[pos]
		# left-adjacent charater of caret
		cursorStr = ''
		if cursorPos and cursorPos < len(composition):
			cursorStr = composition[cursorPos - 1]
		delta = composition[deltaStart:]
		log.io("IM C:%s D:%s S:%s M:%s R:%s P:%d" % (composition, delta, selectedStr, cursorStr, result, cursorPos))
	# if the event comes after candidate selection key, discriminant reading is used.
	if _lastKeyCode in (win32con.VK_SPACE, win32con.VK_UP, win32con.VK_DOWN, win32con.VK_LEFT, win32con.VK_RIGHT):
		if selectedStr:
			candidateUpdated(selectedStr)
		else:
			candidateUpdated(delta)
	else:
		filterInputEvent(composition, delta, selectedStr, cursorStr, result, cursorPos)	

WINFUNC = WINFUNCTYPE(c_void_p,c_uint,c_uint,c_wchar_p,c_wchar_p,c_wchar_p,c_wchar_p)
cmp_func = WINFUNC(inputMethodCallback)

# initialize
def initialize():
	global lib
	try:
		if config.conf["keyboard"]["nvdajpEnableKeyEvents"]:
			lib = windll.LoadLibrary("lib/nvdajpime.dll")
			lib.Initialize(cmp_func)
	except:
		pass

# terminate
def terminate():
	try:
		if lib:
			lib.Terminate()
			windll.FreeLibrary(lib)
	except:
		pass
