# coding: UTF-8
#translator.py (Japanese Braille translator)
#A part of NonVisual Desktop Access (NVDA)
#Copyright (C) 2012 Takuya Nishimoto (NVDA Japanese Team)
#This file is covered by the GNU General Public License.
#See the file COPYING for more details.

import os
import copy
from mecab import *
import tenji

_logwrite = None

try:
	from logHandler import log
	_logwrite = log.debug
except:
	def __print(s): print s
	_logwrite = __print

CONNECTED_MORPHS = {
	u'について': [
		[u'に', u'ニ', '0/1', None, None, '*'],
		[u'ついて', u'ツイテ', '1/3', u'動詞', '*', '*'],
		],
	u'による': [
		[u'に', u'ニ', '0/1', None, None, '*'],
		[u'よる', u'ヨル', '0/2', u'動詞', '*', '*'],
		],
	u'において': [
		[u'に', u'ニ', '0/1', None, None, '*'],
		[u'おいて', u'オイテ', '0/3', u'動詞', '*', '*'],
		],
	u'によって': [
		[u'に', u'ニ', '0/1', None, None, '*'],
		[u'よって', u'ヨッテ', '0/3', u'動詞', '*', '*'],
		],
	u'に対して': [
		[u'に', u'ニ', '0/1', None, None, '*'],
		[u'対して', u'タイシテ', '1/4', u'動詞', '*', '*'],
		],
	u'に関して': [
		[u'に', u'ニ', '0/1', None, None, '*'],
		[u'関して', u'カンシテ', '1/4', u'動詞', '*', '*'],
		],
	u'という': [
		[u'と', u'ト', '0/1', None, None, '*'],
		[u'いう', u'イウ', '0/2', u'動詞', '*', '*'],
		],
}

REPLACE_MORPHS = {
	u'システムキャレット': [
		[u'システム', u'システム', '0/4', None, None, None],
		[u'キャレット', u'キャレット', '1/4', None, None, None],
		],
	u'フィードバック': [
		[u'フィード', u'フィード', '0/3', None, None, None],
		[u'バック', u'バック', '1/3', None, None, None],
		],
	u'インターフェース': [
		[u'インター', u'インター', '0/3', None, None, None],
		[u'フェース', u'フェース', '1/3', None, None, None],
		],
	u'オペレーティングシステム': [
		[u'オペレーティング', u'オペレーティング', '0/7', None, None, None],
		[u'システム', u'システム', '1/4', None, None, None],
		],
	u'トーキングインストーラー': [
		[u'トーキング', u'トーキング', '0/5', None, None, None],
		[u'インストーラー', u'インストーラー', '4/7', None, None, None],
		],
	u'アイスクリーム': [
		[u'アイス', u'アイス', '0/3', None, None, None],
		[u'クリーム', u'クリーム', '2/4', None, None, None],
		],
	u'日本点字図書館': [
		[u'日本', u'ニッポン', '3/4', None, None, None],
		[u'点字', u'テンジ', '0/3', None, None, None],
		[u'図書館', u'トショカン', '2/4', None, None, None],
		],
	u'通り': [
		[u'通り', u'トオリ', '1/3', None, None, None],
		],
	u'ねえ': [
		[u'ねえ', u'ネエ', '1/2', None, None, None],
		],
	u'お母さん': [
		[u'お母さん', u'オカアサン', '2/5', None, None, None],
		],
	u'閏年': [
		[u'閏年', u'ウルウドシ', '2/5', None, None, None],
		],
	u'狼': [
		[u'狼', u'オオカミ', '1/4', None, None, None],
		],
	u'通る': [
		[u'通る', u'トオル', '1/3', None, None, None],
		],
	u'多い': [
		[u'多い', u'オオイ', '1/3', None, None, None],
		],
	u'多く': [
		[u'多く', u'オオク', '1/3', None, None, None],
		],
	u'大晦日': [
		[u'大晦日', u'オオミソカ', '3/5', None, None, None],
		],
	u'手作り': [
		[u'手作り', u'テヅクリ', '2/4', None, None, None],
		],
	u'南半球': [
		[u'南', u'ミナミ', '0/3', None, None, None],
		[u'半球', u'ハンキュー', '1/4', None, None, None],
		],
	u'アメリカ合衆国': [
		[u'アメリカ', u'アメリカ', '0/4', None, None, None],
		[u'合衆国', u'ガッシューコク', '3/6', None, None, None],
		],
	u'第一人者': [
		[u'第一人者', u'ダイ1ニンシャ', '5/7', None, None, None],
		],
	u'一流': [
		[u'一流', u'1リュー', '0/4', None, None, None],
		],
	u'一月': [
		[u'一月', u'1ガツ', '2/4', None, None, None],
		],
	u'四方': [
		[u'四方', u'4ホー', '1/3', None, None, None],
		],
	u'六法全書': [
		[u'六法', u'6ポー', '0/4', None, None, None],
		[u'全書', u'ゼンショ', '1/3', None, None, None],
		],
	u'百人一首': [
		[u'百人', u'100ニン', '0/4', None, None, None],
		[u'一首', u'1シュ', '1/3', None, None, None],
		],
	u'一': [
		[u'一', u'1', '1/2', None, None, None],
		],
	u'二': [
		[u'二', u'2', '1/2', None, None, None],
		],
	u'三': [
		[u'三', u'3', '1/2', None, None, None],
		],
	u'四': [
		[u'四', u'4', '1/2', None, None, None],
		],
	u'五': [
		[u'五', u'5', '1/2', None, None, None],
		],
	u'六': [
		[u'六', u'6', '1/2', None, None, None],
		],
	u'七': [
		[u'七', u'7', '1/2', None, None, None],
		],
	u'八': [
		[u'八', u'8', '1/2', None, None, None],
		],
	u'九': [
		[u'九', u'9', '1/2', None, None, None],
		],
	u'〇': [
		[u'〇', u'9', '1/2', None, None, None],
		],
	u'十': [
		[u'十', u'10', '1/2', None, None, None],
		],
	u'十七': [
		[u'十七', u'17', '1/4', None, None, None],
		],
	u'二十五': [
		[u'二十五', u'25', '1/4', None, None, None],
		],
	u'百': [
		[u'百', u'100', '2/2', None, None, None],
		],
	u'二〇〇〇': [
		[u'二〇〇〇', u'2000', '1/3', None, None, None],
		],
}

class MecabMorph(object):
	__slots__ = ('hyouki', 'nhyouki', 'hinshi1', 'hinshi2', 'hinshi3', 'hinshi4', 
				 'kana', 'yomi', 'accent', 'output', 'sepflag')

	def is_substantive_word(self):
		if self.hinshi1 == u'記号': return False
		if self.hinshi2 == u'接頭': return True
		if self.hinshi2 == u'接尾': return True
		if self.hinshi1 in (u'助動詞', u'助詞'): return True
		return False

	def is_independent_word(self):
		if self.hinshi1 == u'記号': return False
		return not self.is_substantive_word()

	def write(self, logwrite):
		logwrite("%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%d" % 
				 (self.hyouki, self.nhyouki, 
				  self.hinshi1, self.hinshi2, self.hinshi3, self.hinshi4, 
				 self.kana, self.yomi, self.accent, self.output, self.sepflag))

def mecab_to_morphs(mf):
	li = []
	if mf is None or mf.feature is None or mf.size is None: 
		return li
	for i in xrange(0, mf.size):
		s = string_at(mf.feature[i])
		if s:
			s = s.decode(CODE, 'ignore')
			ar = s.split(",")
			mo = MecabMorph()
			mo.hyouki = ar[0]
			mo.nhyouki = unicodedata.normalize('NFKC',ar[0])
			mo.hinshi1 = ar[1]
			mo.hinshi2 = ar[2]
			if len(ar) > 3:
				mo.hinshi3 = ar[3]
				mo.hinshi4 = ar[4]
			else:
				mo.hinshi3 = mo.hinshi4 = None
			if len(ar) > 9:
				mo.kana = ar[8]
				mo.yomi = ar[9]
				mo.accent = ar[10]
				mo.output = ar[9]
			else:
				# 100,名詞,数,*,*,*,*,*
				mo.kana = ar[0]
				mo.yomi = ar[0]
				mo.accent = ''
				mo.output = ar[0]
			# input:
			#  12,名詞,普通名詞
			# output:
			#  12,名詞,数
			ma = re.match('\d+', mo.hyouki)
			if ma and mo.hinshi2 != u'数':
				mo.hinshi2 = u'数'
				mo.output = mo.kana = mo.yomi = mo.hyouki
			mo.sepflag = False
			li.append(mo)
	return li

def normalize_morphs(li):
	new_li = []
	for mo in li:
		if mo.hinshi3 == u'連語':
			if mo.hyouki in CONNECTED_MORPHS.keys():
				new_morphs = CONNECTED_MORPHS[mo.hyouki]
				for i in new_morphs:
					m = copy.deepcopy(mo)
					m.hyouki = m.nhyouki = i[0] # に
					if i[3]: m.hinshi1 = i[3]
					if i[4]: m.hinshi2 = i[4]
					if i[5]: m.hinshi3 = i[5]
					m.output = m.kana = m.yomi = i[1] # ニ
					m.accent = i[2] # 0/1
					new_li.append(m)
			else:
				new_li.append(mo)
		else:
			new_li.append(mo)
	return new_li

def replace_morphs(li):
	new_li = []
	for mo in li:
		if mo.hyouki in REPLACE_MORPHS.keys():
			new_morphs = REPLACE_MORPHS[mo.hyouki]
			for i in new_morphs:
				m = copy.deepcopy(mo)
				m.hyouki = m.nhyouki = i[0]
				if i[3]: m.hinshi1 = i[3]
				if i[4]: m.hinshi2 = i[4]
				if i[5]: m.hinshi3 = i[5]
				m.output = m.kana = m.yomi = i[1]
				m.accent = i[2]
				new_li.append(m)
		else:
			new_li.append(mo)
	return new_li

def concatinate_num_morphs(li):
	mo = copy.deepcopy(li[0])
	s = ''
	y = ''
	for i in li:
		s += i.hyouki
		y += i.yomi
	mo.hyouki = mo.nhyouki = s
	mo.yomi = mo.kana = mo.output = y
	return mo

def replace_digit_morphs(li):
	# input:
	#  十,名詞,数
	#  七,名詞,数
	# output:
	#  十七,名詞,数
	# input:
	#  二,名詞,数
	#  十,名詞,数
	#  五,名詞,数
	# output:
	#  二十五,名詞,数
	new_li = []
	num_morphs = []
	for mo in li:
		if mo.hinshi2 == u'数':
			num_morphs.append(mo)
		else:
			if len(num_morphs):
				new_li.append(concatinate_num_morphs(num_morphs))
				num_morphs = []
			new_li.append(mo)
	if len(num_morphs):
		new_li.append(concatinate_num_morphs(num_morphs))
	return new_li

def should_separate(prev_mo, mo):
	if prev_mo.hinshi1 == u'助動詞' and prev_mo.hyouki == u'で':
		if mo.hinshi1 == u'助動詞' and mo.hyouki == u'なく': return True
		if mo.hinshi1 == u'助動詞' and mo.hyouki == u'ある': return True
	if prev_mo.hinshi1 == u'接頭詞' and prev_mo.hyouki == u'超' and mo.hinshi1 == u'名詞': return True
	if prev_mo.hinshi1 == u'接頭詞' and mo.hinshi1 == u'名詞': return False
	if prev_mo.hinshi1 == u'助詞' and mo.hinshi1 == u'接頭詞': return True
	if prev_mo.hinshi1 == u'名詞' and prev_mo.hinshi2 == u'ナイ形容詞語幹' and mo.hyouki == u'ない': return True
	if prev_mo.hinshi1 == u'名詞' and mo.hinshi1 == u'名詞':
		if mo.hinshi2 == u'数': return False
		if mo.hinshi2 != u'接尾': return True
	if prev_mo.hinshi1 == u'名詞' and prev_mo.hinshi2 == u'接尾':
		if mo.hinshi1 == u'動詞' and mo.hinshi2 == u'非自立': return False
	if prev_mo.hinshi1 == u'動詞' and prev_mo.hinshi2 == u'自立' and mo.hinshi1 == u'助詞' and mo.hinshi2 == u'接続助詞' and mo.hyouki == u'および': return True
	if prev_mo.hinshi1 == u'動詞' and prev_mo.hinshi2 == u'自立' and mo.hinshi1 == u'動詞' and mo.hinshi2 == u'非自立': return False
	if prev_mo.is_substantive_word() and mo.is_independent_word(): return True
	if prev_mo.is_independent_word() and mo.is_independent_word(): return True
	return False

def is_num_char(c):
	if c in string.digits: return True
	return False

def is_alphabet(c):
	if c in string.ascii_letters: return True
	return False

def is_uppercase(c):
	if c in string.ascii_uppercase: return True
	return False

def insert_space_after_alphabet(text):
	s1 = ''
	after_alphabet = False
	for c in text:
		if is_alphabet(c):
			after_alphabet = True
		else:
			if after_alphabet and c != ' ' and c != ')':
				# non-alphabet after alphabet
				s1 += u' '
			after_alphabet = False
		s1 += c
	return s1

def morphs_to_string(li):
	s = u''
	for i in xrange(0, len(li)):
		s += li[i].output
		if li[i].sepflag: s += u' '
	return s.strip()

def japanese_braille_separate(text, logwrite):
	text = unicodedata.normalize('NFKC',text)
	text = text.encode('utf-8', 'ignore')

	mf = MecabFeatures()
	Mecab_analysis(text, mf)
	Mecab_correctFeatures(mf)
	Mecab_print(mf, logwrite, output_header = False)
	li = mecab_to_morphs(mf)
	mf = None

	for mo in li:
		if mo.hinshi1 == u'空白':
			mo.output = ' '
		elif mo.hyouki in ('.', '-', '(', ')', '*', '/', '+') and mo.hinshi1 == u'名詞':
			mo.hinshi1 = u'記号'
			mo.hinshi2 = '*'
		elif mo.hinshi1 == u'数' and mo.hyouki != mo.nhyouki:
			# digit numbers (not kanji characters)
			mo.output = mo.nhyouki
			# FIXME

	li = normalize_morphs(li)
	li = replace_digit_morphs(li)
	li = replace_morphs(li)

	for i in xrange(1, len(li)):
		li[i-1].sepflag = should_separate(li[i-1], li[i])

	for mo in li:
		mo.write(logwrite)
	logwrite('')

	text = morphs_to_string(li)
	return text

mecab_initialized = False

def initialize(jtalk_dir=None, logwrite=_logwrite):
	global mecab_initialized
	if jtalk_dir:
		Mecab_initialize(logwrite, jtalk_dir)
	else:
		Mecab_initialize(logwrite)
	if logwrite: logwrite("initialize() done.")
	mecab_initialized = True

def terminate():
	global _logwrite
	if _logwrite: _logwrite("terminate() done.")
	global mecab_initialized
	mecab_initialized = False

def separate(text):
	if not mecab_initialized:
		initialize()
	return japanese_braille_separate(text, _logwrite)

# returns u'\u2801\u2802\u2803\u2804\u2805\u2806\u2807'
def japaneseToUnicodeBraille(text):
	sp = separate(text)
	return tenji.tenji(sp).replace(u'□', ' ')

# for brailleViewer
def getReadingAndBraille(text):
	sp = separate(text)
	text = tenji.tenji(sp).replace(u'□', ' ')
	return (sp, text)
